@@ -13,8 +13,10 @@ use std::io::Write;
 use std::sync::mpsc;
 use std::time::Duration;
 
-use uucore::error::UResult;
-use uucore::memo::sprintf;
+use uucore::{
+    error::UResult,
+    format::num_format::{FloatVariant, Formatter},
+};
 
 use crate::numbers::{to_magnitude_and_suffix, SuffixType};
 
@@ -152,7 +154,14 @@ impl ProgUpdate {
         let (carriage_return, newline) = if rewrite { ("\r", "") } else { ("", "\n") };
 
         // The duration should be formatted as in `printf %g`.
-        let duration_str = sprintf("%g", &[duration.to_string()])?;
+        let mut duration_str = Vec::new();
+        uucore::format::num_format::Float {
+            variant: FloatVariant::Shortest,
+            ..Default::default()
+        }
+        .fmt(&mut duration_str, duration)?;
+        // We assume that printf will output valid UTF-8
+        let duration_str = std::str::from_utf8(&duration_str).unwrap();
 
         // If the number of bytes written is sufficiently large, then
         // print a more concise representation of the number, like
@@ -379,6 +388,17 @@ impl std::ops::AddAssign for WriteStat {
     }
 }
 
+impl std::ops::Add for WriteStat {
+    type Output = Self;
+    fn add(self, other: Self) -> Self {
+        Self {
+            writes_complete: self.writes_complete + other.writes_complete,
+            writes_partial: self.writes_partial + other.writes_partial,
+            bytes_total: self.bytes_total + other.bytes_total,
+        }
+    }
+}
+
 /// How much detail to report when printing transfer statistics.
 ///
 /// This corresponds to the available settings of the `status`
