@@ -1,9 +1,7 @@
-// * This file is part of the uutils coreutils package.
-// *
-// * (c) 2020 nicoo <nicoo@debian.org>
-// *
-// * For the full copyright and license information, please view the LICENSE file
-// * that was distributed with this source code.
+// This file is part of the uutils coreutils package.
+//
+// For the full copyright and license information, please view the LICENSE
+// file that was distributed with this source code.
 
 // spell-checker:ignore (URL) appspot
 
@@ -16,19 +14,17 @@ pub(crate) trait Basis {
 impl Basis for Montgomery<u64> {
     // Small set of bases for the Miller-Rabin prime test, valid for all 64b integers;
     //  discovered by Jim Sinclair on 2011-04-20, see miller-rabin.appspot.com
-    #[allow(clippy::unreadable_literal)]
-    const BASIS: &'static [u64] = &[2, 325, 9375, 28178, 450775, 9780504, 1795265022];
+    const BASIS: &'static [u64] = &[2, 325, 9375, 28178, 450_775, 9_780_504, 1_795_265_022];
 }
 
 impl Basis for Montgomery<u32> {
     // spell-checker:ignore (names) Steve Worley
     // Small set of bases for the Miller-Rabin prime test, valid for all 32b integers;
     //  discovered by Steve Worley on 2013-05-27, see miller-rabin.appspot.com
-    #[allow(clippy::unreadable_literal)]
     const BASIS: &'static [u64] = &[
-        4230279247111683200,
-        14694767155120705706,
-        16641139526367750375,
+        4_230_279_247_111_683_200,
+        14_694_767_155_120_705_706,
+        16_641_139_526_367_750_375,
     ];
 }
 
@@ -63,13 +59,13 @@ pub(crate) fn test<A: Arithmetic + Basis>(m: A) -> Result {
     let one = m.one();
     let minus_one = m.minus_one();
 
-    'witness: for _a in A::BASIS.iter() {
+    'witness: for _a in A::BASIS {
         let _a = _a % n;
         if _a == 0 {
             continue;
         }
 
-        let a = m.from_u64(_a);
+        let a = m.to_mod(_a);
 
         // x = a^r mod n
         let mut x = m.pow(a, r);
@@ -112,10 +108,9 @@ pub(crate) fn is_prime(n: u64) -> bool {
 mod tests {
     use super::*;
     use crate::numeric::{traits::DoubleInt, Arithmetic, Montgomery};
-    use crate::parametrized_check;
     use quickcheck::quickcheck;
     use std::iter;
-    const LARGEST_U64_PRIME: u64 = 0xFFFFFFFFFFFFFFC5;
+    const LARGEST_U64_PRIME: u64 = 0xFFFF_FFFF_FFFF_FFC5;
 
     fn primes() -> impl Iterator<Item = u64> {
         iter::once(2).chain(odd_primes())
@@ -153,12 +148,20 @@ mod tests {
         for p in odd_primes() {
             assert!(
                 test(Montgomery::<A>::new(p)).is_prime(),
-                "{} reported composite",
-                p
+                "{p} reported composite"
             );
         }
     }
-    parametrized_check!(first_primes);
+
+    #[test]
+    fn first_primes_u32() {
+        first_primes::<u32>();
+    }
+
+    #[test]
+    fn first_primes_u64() {
+        first_primes::<u64>();
+    }
 
     #[test]
     fn one() {
@@ -173,7 +176,7 @@ mod tests {
     fn first_composites() {
         for (p, q) in primes().zip(odd_primes()) {
             for i in p + 1..q {
-                assert!(!is_prime(i), "{} reported prime", i);
+                assert!(!is_prime(i), "{i} reported prime");
             }
         }
     }
@@ -192,11 +195,20 @@ mod tests {
             for q in odd_primes().take_while(|q| *q <= p) {
                 let n = p * q;
                 let m = Montgomery::<A>::new(n);
-                assert!(!test(m).is_prime(), "{} = {} × {} reported prime", n, p, q);
+                assert!(!test(m).is_prime(), "{n} = {p} × {q} reported prime");
             }
         }
     }
-    parametrized_check!(small_semiprimes);
+
+    #[test]
+    fn small_semiprimes_u32() {
+        small_semiprimes::<u32>();
+    }
+
+    #[test]
+    fn small_semiprimes_u64() {
+        small_semiprimes::<u64>();
+    }
 
     quickcheck! {
         fn composites(i: u64, j: u64) -> bool {
