@@ -1,7 +1,11 @@
+// This file is part of the uutils coreutils package.
+//
+// For the full copyright and license information, please view the LICENSE
+// file that was distributed with this source code.
 use crate::common::util::TestScenario;
 use regex::Regex;
 #[cfg(all(unix, not(target_os = "macos")))]
-use rust_users::get_effective_uid;
+use uucore::process::geteuid;
 
 #[test]
 fn test_invalid_arg() {
@@ -213,7 +217,7 @@ fn test_date_format_literal() {
 #[test]
 #[cfg(all(unix, not(target_os = "macos")))]
 fn test_date_set_valid() {
-    if get_effective_uid() == 0 {
+    if geteuid() == 0 {
         new_ucmd!()
             .arg("--set")
             .arg("2020-03-12 13:30:00+08:00")
@@ -234,7 +238,7 @@ fn test_date_set_invalid() {
 #[test]
 #[cfg(all(unix, not(any(target_os = "android", target_os = "macos"))))]
 fn test_date_set_permissions_error() {
-    if !(get_effective_uid() == 0 || uucore::os::is_wsl_1()) {
+    if !(geteuid() == 0 || uucore::os::is_wsl_1()) {
         let result = new_ucmd!()
             .arg("--set")
             .arg("2020-03-11 21:45:00+08:00")
@@ -261,7 +265,7 @@ fn test_date_set_mac_unavailable() {
 #[cfg(all(unix, not(target_os = "macos")))]
 /// TODO: expected to fail currently; change to succeeds() when required.
 fn test_date_set_valid_2() {
-    if get_effective_uid() == 0 {
+    if geteuid() == 0 {
         let result = new_ucmd!()
             .arg("--set")
             .arg("Sat 20 Mar 2021 14:53:01 AWST") // spell-checker:disable-line
@@ -290,6 +294,7 @@ fn test_date_for_no_permission_file() {
     use std::os::unix::fs::PermissionsExt;
     let file = std::fs::OpenOptions::new()
         .create(true)
+        .truncate(true)
         .write(true)
         .open(at.plus(FILE))
         .unwrap();
@@ -325,7 +330,7 @@ fn test_date_for_file() {
 #[cfg(all(unix, not(target_os = "macos")))]
 /// TODO: expected to fail currently; change to succeeds() when required.
 fn test_date_set_valid_3() {
-    if get_effective_uid() == 0 {
+    if geteuid() == 0 {
         let result = new_ucmd!()
             .arg("--set")
             .arg("Sat 20 Mar 2021 14:53:01") // Local timezone
@@ -339,7 +344,7 @@ fn test_date_set_valid_3() {
 #[cfg(all(unix, not(target_os = "macos")))]
 /// TODO: expected to fail currently; change to succeeds() when required.
 fn test_date_set_valid_4() {
-    if get_effective_uid() == 0 {
+    if geteuid() == 0 {
         let result = new_ucmd!()
             .arg("--set")
             .arg("2020-03-11 21:45:00") // Local timezone
@@ -395,3 +400,12 @@ fn test_invalid_date_string() {
         .no_stdout()
         .stderr_contains("invalid date");
 }
+
+#[test]
+fn test_date_overflow() {
+    new_ucmd!()
+        .arg("-d68888888888888sms")
+        .fails()
+        .no_stdout()
+        .stderr_contains("invalid date");
+}
