@@ -1,3 +1,7 @@
+// This file is part of the uutils coreutils package.
+//
+// For the full copyright and license information, please view the LICENSE
+// file that was distributed with this source code.
 // spell-checker:ignore (paths) gnutest
 
 use crate::common::util::TestScenario;
@@ -662,12 +666,88 @@ fn test_invalid_stdin_number_returns_status_2() {
 
 #[test]
 fn test_invalid_stdin_number_in_middle_of_input() {
-    new_ucmd!().pipe_in("100\nhello\n200").fails().code_is(2);
+    new_ucmd!()
+        .pipe_in("100\nhello\n200")
+        .ignore_stdin_write_error()
+        .fails()
+        .stdout_is("100\n")
+        .code_is(2);
 }
 
 #[test]
-fn test_invalid_argument_number_returns_status_2() {
-    new_ucmd!().args(&["hello"]).fails().code_is(2);
+fn test_invalid_stdin_number_with_warn_returns_status_0() {
+    new_ucmd!()
+        .args(&["--invalid=warn"])
+        .pipe_in("4Q")
+        .succeeds()
+        .stdout_is("4Q\n")
+        .stderr_is("numfmt: invalid suffix in input: '4Q'\n");
+}
+
+#[test]
+fn test_invalid_stdin_number_with_ignore_returns_status_0() {
+    new_ucmd!()
+        .args(&["--invalid=ignore"])
+        .pipe_in("4Q")
+        .succeeds()
+        .stdout_only("4Q\n");
+}
+
+#[test]
+fn test_invalid_stdin_number_with_abort_returns_status_2() {
+    new_ucmd!()
+        .args(&["--invalid=abort"])
+        .pipe_in("4Q")
+        .fails()
+        .code_is(2)
+        .stderr_only("numfmt: invalid suffix in input: '4Q'\n");
+}
+
+#[test]
+fn test_invalid_stdin_number_with_fail_returns_status_2() {
+    new_ucmd!()
+        .args(&["--invalid=fail"])
+        .pipe_in("4Q")
+        .fails()
+        .code_is(2)
+        .stdout_is("4Q\n")
+        .stderr_is("numfmt: invalid suffix in input: '4Q'\n");
+}
+
+#[test]
+fn test_invalid_arg_number_with_warn_returns_status_0() {
+    new_ucmd!()
+        .args(&["--invalid=warn", "4Q"])
+        .succeeds()
+        .stdout_is("4Q\n")
+        .stderr_is("numfmt: invalid suffix in input: '4Q'\n");
+}
+
+#[test]
+fn test_invalid_arg_number_with_ignore_returns_status_0() {
+    new_ucmd!()
+        .args(&["--invalid=ignore", "4Q"])
+        .succeeds()
+        .stdout_only("4Q\n");
+}
+
+#[test]
+fn test_invalid_arg_number_with_abort_returns_status_2() {
+    new_ucmd!()
+        .args(&["--invalid=abort", "4Q"])
+        .fails()
+        .code_is(2)
+        .stderr_only("numfmt: invalid suffix in input: '4Q'\n");
+}
+
+#[test]
+fn test_invalid_arg_number_with_fail_returns_status_2() {
+    new_ucmd!()
+        .args(&["--invalid=fail", "4Q"])
+        .fails()
+        .code_is(2)
+        .stdout_is("4Q\n")
+        .stderr_is("numfmt: invalid suffix in input: '4Q'\n");
 }
 
 #[test]
